"""
Abstractions for lazy compositions/manipulations of And Inverter
Graphs.
"""

from __future__ import annotations

from typing import (Union, FrozenSet, Callable, Tuple,
                    Mapping, Sequence, Optional)

import attr
import funcy as fn
from bidict import bidict
from pyrsistent import pmap
from pyrsistent.typing import PMap

import aiger as A
from aiger.aig import AIG, Node, Input, LatchIn
from aiger.aig import ConstFalse


@attr.s(frozen=True, auto_attribs=True)
class LazyAIG:
    def __call__(self, inputs, latches=None, *, lift=None):
        pass

    @property
    def latch2init(self):
        pass

    @property
    def inputs(self):
        pass

    @property
    def outputs(self):
        pass

    @property
    def comments(self):
        pass

    def write(self, path):
        self.aig.write(path)

    relabel = AIG.relabel
    simulator = AIG.simulator
    simulate = AIG.simulate

    @property
    def latches(self) -> FrozenSet[str]:
        return frozenset(self.latch2init.keys())

    @property
    def lazy_aig(self) -> LazyAIG:
        return self

    @property
    def aig(self) -> AIG:
        """Return's flattened AIG represented by this LazyAIG."""
        false = ConstFalse()
        inputs = {i: Input(i) for i in self.inputs}
        latches = {i: LatchIn(i) for i in self.latches}

        def lift(obj):
            if isinstance(obj, Node):
                return obj
            assert isinstance(obj, bool)
            return ~false if obj else false

        node_map, latch_map = self(inputs, latches=latches, lift=lift)
        return AIG(
            comments=self.comments,
            inputs=self.inputs,
            node_map=node_map,
            latch_map=latch_map,
            latch2init=self.latch2init,
        )

    def __rshift__(self, other: AIG_Like) -> LazyAIG:
        """Cascading composition. Feeds self into other."""
        return Cascading(self, other)

    def __lshift__(self, other: AIG_Like) -> LazyAIG:
        """Cascading composition. Feeds other into self."""
        return lazy(other) >> self

    def __or__(self, other: AIG_Like) -> LazyAIG:
        """Parallel composition between self and other."""
        assert not self.latches & other.latches
        assert not self.outputs & other.outputs
        return Parallel(self, other)

    def cutlatches(self, latches=None, renamer=None) -> Tuple[LazyAIG, Labels]:
        """Returns LazyAIG where the latches specified
        in `latches` have been converted into inputs/outputs.

        - If `latches` is `None`, then all latches are cut.
        - `renamer`: is a function from strings to strings which
           determines how to rename latches to avoid name collisions.
        """
        lcirc = CutLatches(self, renamer=renamer, cut=latches)
        l2init = dict(self.latch2init)
        lmap = {k: (lcirc.renamer(k), l2init[k]) for k in lcirc.cut}
        return lcirc, lmap

    def loopback(self, *wirings) -> LazyAIG:
        """Returns result of feeding outputs specified in `*wirings` to
        inputs specified in `wirings`.

        Each positional argument (element of wirings) should have the following
        schema:

           {
              'input': str,
              'output': str,
              'latch': str,         # what to name the new latch.
              'init': bool,         # new latch's initial value.
              'keep_output': bool,  # whether output is consumed by feedback.
            }
        """
        return LoopBack(self, wirings=wirings)

    def unroll(self, horizon, *, init=True, omit_latches=True,
               only_last_outputs=False) -> LazyAIG:
        """
        Returns circuit which computes the same function as
        the sequential circuit after `horizon` many inputs.

        Each input/output has `##time_{time}` appended to it to
        distinguish different time steps.
        """
        return A.Unrolled(
            self, horizon, init, omit_latches, only_last_outputs
        )

    def __getitem__(self, others):
        """Relabel inputs, outputs, or latches.

        `others` is a tuple, (kind, relabels), where

          1. kind in {'i', 'o', 'l'}
          2. relabels is a mapping from old names to new names.

        Note: The syntax is meant to resemble variable substitution
        notations, i.e., foo[x <- y] or foo[x / y].
        """
        assert isinstance(others, tuple) and len(others) == 2
        kind, relabels = others
        assert kind in {'i', 'o', 'l'}
        key = {
            'i': 'input_relabels',
            'l': 'latch_relabels',
            'o': 'output_relabels',
        }.get(kind)

        return A.Relabeled(self, **{key: relabels})

    def reinit(self, latch2init) -> LazyAIG:
        """Update late initial values based on mapping provided."""
        assert set(latch2init.keys()) <= self.latches
        return UpdatedLatchInits(circ=self, latch2init=latch2init)


AIG_Like = Union[AIG, LazyAIG]
Labels = Mapping[str, str]


def walk_keys(func, mapping):
    return fn.walk_keys(func, dict(mapping))


def omit(mapping, keys):
    return fn.omit(dict(mapping), keys)


def project(mapping, keys):
    return fn.project(dict(mapping), keys)


@attr.s(frozen=True, auto_attribs=True)
class Parallel(LazyAIG):
    left: AIG_Like
    right: AIG_Like

    def __call__(self, inputs, latches=None, *, lift=None):
        out_l, lmap_l = self.left(inputs, latches=latches, lift=lift)
        out_r, lmap_r = self.right(inputs, latches=latches, lift=lift)
        return fn.merge(out_l, out_r), fn.merge(lmap_l, lmap_r)

    def _merge_maps(self, key):
        map1, map2 = [pmap(getattr(c, key)) for c in (self.left, self.right)]
        return map1 + map2

    @property
    def latch2init(self):
        return self._merge_maps('latch2init')

    @property
    def inputs(self):
        return self.left.inputs | self.right.inputs

    @property
    def outputs(self):
        return self.left.outputs | self.right.outputs

    @property
    def comments(self):
        return self.left.comments + self.right.comments


@attr.s(frozen=True, auto_attribs=True)
class Wire:
    input: str
    output: str
    latch: str
    keep_output: bool = True
    init: bool = True


def convert_wirings(wirings):
    for wire in wirings:
        wire.setdefault('latch', wire['input'])

    return tuple(Wire(**w) for w in wirings)


@attr.s(frozen=True, auto_attribs=True)
class LoopBack(LazyAIG):
    circ: AIG_Like
    wirings: Sequence[Wire] = attr.ib(converter=convert_wirings)

    def __call__(self, inputs, latches=None, *, lift=None):
        if latches is None:
            latches = pmap()
        latches = dict(self.latch2init + latches)  # Override initial values.

        for wire in self.wirings:
            inputs[wire.input] = latches[wire.latch]
            del latches[wire.latch]

        omap, lmap = self.circ(inputs, latches=latches, lift=lift)

        for wire in self.wirings:
            out, latch = wire.output, wire.latch
            lmap[latch] = omap[out]
            if not wire.keep_output:
                del omap[out]

        return omap, lmap

    @property
    def latch2init(self):
        latch2init = pmap(self.circ.latch2init).evolver()
        for wire in self.wirings:
            latch2init[wire.latch] = wire.init
        return latch2init.persistent()

    @property
    def inputs(self):
        return self.circ.inputs - {w.input for w in self.wirings}

    @property
    def outputs(self):
        omitted = {w.output for w in self.wirings if not w.keep_output}
        return self.circ.outputs - omitted

    @property
    def comments(self):
        return self.circ.comments


def convert_renamer(renamer):
    if renamer is None:
        def renamer(*_):
            return A.common._fresh()
    return fn.memoize(renamer)


@attr.s(frozen=True, auto_attribs=True)
class CutLatches(LazyAIG):
    circ: AIG_Like
    renamer: Callable[[str], str] = attr.ib(converter=convert_renamer)
    cut: Optional[FrozenSet[str]] = None

    def __attrs_post_init__(self):
        if self.cut is None:
            object.__setattr__(self, "cut", self.circ.latches)

    def __call__(self, inputs, latches=None, *, lift=None):
        inputs = dict(inputs)
        if latches is None:
            latches = pmap()
        latches = dict(self.latch2init + latches)  # Override initial values.

        for latch in self.cut:
            new_name = self.renamer(latch)
            latches[latch] = inputs[new_name]
            del inputs[new_name]

        omap, lmap = self.circ(inputs, latches=latches, lift=lift)

        for latch in self.cut:
            new_name = self.renamer(latch)
            omap[new_name] = lmap[latch]
            del lmap[latch]

        return omap, lmap

    @property
    def latch2init(self):
        return pmap(omit(self.circ.latch2init, self.cut))

    @property
    def inputs(self):
        return self.circ.inputs | set(map(self.renamer, self.cut))

    @property
    def outputs(self):
        return self.circ.outputs | set(map(self.renamer, self.cut))

    @property
    def comments(self):
        return self.circ.comments


@attr.s(frozen=True, auto_attribs=True)
class Cascading(LazyAIG):
    left: AIG_Like
    right: AIG_Like

    def __call__(self, inputs, latches=None, *, lift=None):
        inputs_l = project(inputs, self.left.inputs)
        omap_l, lmap_l = self.left(inputs_l, latches=latches, lift=lift)

        inputs_r = project(inputs, self.right.inputs)
        inputs_r.update(omap_l)  # <--- Cascade setup happens here.
        omap_l = omit(omap_l, self._interface)

        omap_r, lmap_r = self.right(inputs_r, latches=latches, lift=lift)
        return fn.merge(omap_l, omap_r), fn.merge(lmap_l, lmap_r)

    def _merge_maps(self, key):
        map1, map2 = [pmap(getattr(c, key)) for c in (self.left, self.right)]
        return map1 + map2

    @property
    def latch2init(self):
        return self._merge_maps('latch2init')

    @property
    def _interface(self):
        return self.left.outputs & self.right.inputs

    @property
    def inputs(self):
        return self.left.inputs | (self.right.inputs - self._interface)

    @property
    def outputs(self):
        return self.right.outputs | (self.left.outputs - self._interface)

    @property
    def comments(self):
        return self.left.comments + self.right.comments


def _relabel_map(relabels, mapping):
    return pmap(walk_keys(lambda x: relabels.get(x, x), mapping))


@attr.s(frozen=True, auto_attribs=True)
class UpdatedLatchInits(LazyAIG):
    circ: AIG_Like
    _latch2init: PMap[str, bool] = pmap()

    def __call__(self, inputs, latches=None, *, lift=None):
        if latches is None:
            latches = pmap()

        latches = dict(self.latch2init + latches)  # Override initial values.
        return self.circ(inputs, latches=latches, lift=lift)

    @property
    def latch2init(self):
        return self.circ.latch2init + self._latch2init

    @property
    def inputs(self):
        return self.circ.inputs

    @property
    def outputs(self):
        return self.circ.outputs

    @property
    def comments(self):
        return self.circ.comments


@attr.s(frozen=True, auto_attribs=True)
class Relabeled(LazyAIG):
    circ: AIG_Like
    input_relabels: PMap[str, str] = pmap()
    latch_relabels: PMap[str, str] = pmap()
    output_relabels: PMap[str, str] = pmap()

    def __call__(self, inputs, latches=None, *, lift=None):
        if latches is None:
            latches = pmap()

        latches = dict(self.latch2init + latches)  # Override initial values.

        new2old_i = bidict(self.input_relabels).inv
        new2old_l = bidict(self.latch_relabels).inv
        inputs = _relabel_map(new2old_i, inputs)
        latches = _relabel_map(new2old_l, latches)

        omap, lmap = self.circ(inputs, latches=latches, lift=lift)

        omap = _relabel_map(self.output_relabels, omap)
        lmap = _relabel_map(self.latch_relabels, lmap)
        return dict(omap), dict(lmap)

    @property
    def latch2init(self):
        return _relabel_map(self.latch_relabels, self.circ.latch2init)

    @property
    def inputs(self):
        old_inputs = self.circ.inputs
        return frozenset(self.input_relabels.get(i, i) for i in old_inputs)

    @property
    def outputs(self):
        old_outputs = self.circ.outputs
        return frozenset(self.output_relabels.get(i, i) for i in old_outputs)

    @property
    def comments(self):
        return self.circ.comments


@attr.s(frozen=True, auto_attribs=True)
class Unrolled(LazyAIG):
    circ: AIG_Like
    horizon: int
    init: bool = True
    omit_latches: bool = True
    only_last_outputs: bool = False

    def __call__(self, inputs, latches=None, *, lift=None):
        circ, omit_latches, init = self.circ, self.omit_latches, self.init
        horizon, only_last_outputs = self.horizon, self.only_last_outputs

        if not omit_latches:
            assert (circ.latches & circ.outputs) == set()

        if not init:
            assert (circ.latches & circ.inputs) == set()

        latches = circ.latch2init if init else project(inputs, circ.inputs)
        if init:
            inputs = omit(inputs, circ.inputs)

        outputs = {}
        for time in range(horizon):
            omap, latches = circ(
                inputs={i: inputs[f'{i}##time_{time}'] for i in circ.inputs},
                latches=latches,
                lift=lift
            )

            if (not only_last_outputs) or (time + 1 == horizon):
                template = '{}' + f'##time_{time + 1}'
                outputs.update(walk_keys(template.format, omap))

                if not self.omit_latches:
                    outputs.update(walk_keys(template.format, latches))

        assert set(outputs.keys()) == self.outputs

        return dict(outputs), {}

    @property
    def latch2init(self):
        return pmap()

    def __with_times(self, keys, times):
        for time in times:
            template = '{}' + f'##time_{time}'
            yield from map(template.format, keys)

    def _with_times(self, keys, times):
        return frozenset(self.__with_times(keys, times))

    @property
    def inputs(self):
        base = set() if self.init else self.circ.latches
        base |= self.circ.inputs
        return self._with_times(base, times=range(self.horizon))

    @property
    def outputs(self):
        start = self.horizon if self.only_last_outputs else 1
        base = set() if self.omit_latches else self.circ.latches
        base |= self.circ.outputs
        return self._with_times(base, times=range(start, self.horizon + 1))

    @property
    def comments(self):
        return self.circ.comments


@attr.s(frozen=True, auto_attribs=True)
class Lifted(LazyAIG):
    circ: AIG_Like

    def __call__(self, inputs, latches=None, *, lift=None):
        return self.circ(inputs=inputs, latches=latches, lift=lift)

    @property
    def latch2init(self):
        return self.circ.latch2init

    @property
    def inputs(self):
        return self.circ.inputs

    @property
    def outputs(self):
        return self.circ.outputs

    @property
    def comments(self):
        return self.circ.comments


def lazy(circ: Union[AIG, LazyAIG]) -> LazyAIG:
    """Lifts AIG to a LazyAIG."""
    return Lifted(circ)


__all__ = ['lazy', 'LazyAIG', 'Parallel', 'LoopBack', 'CutLatches',
           'Cascading', 'Relabeled', 'Unrolled', 'AIG_Like']



import io
import re
from collections import defaultdict
from functools import reduce
from typing import Mapping, List, Optional

import attr
import funcy as fn
from bidict import bidict
from toposort import toposort_flatten
from uuid import uuid1
from sortedcontainers import SortedSet, SortedDict

import aiger as A


@attr.s(auto_attribs=True, repr=False)
class Header:
    max_var_index: int
    num_inputs: int
    num_latches: int
    num_outputs: int
    num_ands: int

    def __repr__(self):
        return f"aag {self.max_var_index} {self.num_inputs} " \
            f"{self.num_latches} {self.num_outputs} {self.num_ands}"


NOT_DONE_PARSING_ERROR = "Lines exhausted before parsing ended!\n{}"
DONE_PARSING_ERROR = "Lines exhausted before parsing ended!\n{}"


@attr.s(auto_attribs=True, frozen=True)
class Latch:
    id: int
    input: int
    init: bool = attr.ib(converter=bool)


@attr.s(auto_attribs=True, frozen=True)
class Symbol:
    kind: str
    name: str
    index: int


def fresh():
    return str(uuid1())


@attr.s(auto_attribs=True, frozen=True)
class SymbolTable:
    inputs: Mapping[int, str] = attr.ib(factory=lambda: defaultdict(fresh))
    outputs: Mapping[int, str] = attr.ib(factory=lambda: defaultdict(fresh))
    latches: Mapping[int, str] = attr.ib(factory=lambda: defaultdict(fresh))


@attr.s(auto_attribs=True)
class State:
    header: Optional[Header] = None
    inputs: SortedSet = attr.ib(factory=SortedSet)
    outputs: SortedSet = attr.ib(factory=SortedSet)
    latches: SortedSet = attr.ib(factory=SortedSet)
    symbols: SymbolTable = attr.ib(factory=SymbolTable)
    comments: Optional[List[str]] = None
    nodes: SortedDict = attr.ib(factory=SortedDict)

    @property
    def remaining_latches(self):
        return self.header.num_latches - len(self.latches)

    @property
    def remaining_outputs(self):
        return self.header.num_outputs - len(self.outputs)

    @property
    def remaining_inputs(self):
        return self.header.num_inputs - len(self.inputs)


HEADER_PATTERN = re.compile(r"aag (\d+) (\d+) (\d+) (\d+) (\d+)\n")


def parse_header(state, line) -> bool:
    if state.header is not None:
        return False

    match = HEADER_PATTERN.match(line)
    if not match:
        raise ValueError(f"Failed to parse aag HEADER. {line}")

    try:
        ids = fn.lmap(int, match.groups())

        if any(x < 0 for x in ids):
            raise ValueError("Indicies must be positive!")

        max_idx, nin, nlatch, nout, nand = ids
        if nin + nlatch + nand > max_idx:
            raise ValueError("Sum of claimed indices greater than max.")

        state.header = Header(
            max_var_index=max_idx,
            num_inputs=nin,
            num_latches=nlatch,
            num_outputs=nout,
            num_ands=nand,
        )

    except ValueError as err:
        raise ValueError(f"Failed parsing the header: {err}")
    return True


IO_PATTERN = re.compile(r"(\d+)\n")


def parse_input(state, line) -> bool:
    match = IO_PATTERN.match(line)
    if match is None or state.remaining_inputs <= 0:
        return False
    lit = int(line)
    state.inputs.add(lit)
    state.nodes[lit] = set()
    return True


def parse_output(state, line) -> bool:
    match = IO_PATTERN.match(line)
    if match is None or state.remaining_outputs <= 0:
        return False

    lit = int(line)
    state.outputs.add(lit)
    if lit & 1:
        state.nodes[lit] = {lit ^ 1}
    return True


LATCH_PATTERN = re.compile(r"(\d+) (\d+)(?: (\d+))?\n")


def parse_latch(state, line) -> bool:
    if state.remaining_latches <= 0:
        return False

    match = LATCH_PATTERN.match(line)
    if match is None:
        raise ValueError("Expecting a latch: {line}")

    elems = match.groups()
    if elems[2] is None:
        elems = elems[:2] + (0,)
    elems = fn.lmap(int, elems)

    state.latches.add(Latch(*elems))
    state.nodes[elems[0]] = set()  # Add latch in as source.
    lit = elems[1]  # Add latchout to dependency graph.
    if lit & 1:
        state.nodes[lit] = {lit ^ 1}
    return True


AND_PATTERN = re.compile(r"(\d+) (\d+) (\d+)\n")


def parse_and(state, line) -> bool:
    if state.header.num_ands <= 0:
        return False

    match = AND_PATTERN.match(line)
    if match is None:
        return False

    elems = fn.lmap(int, match.groups())
    state.header.num_ands -= 1
    deps = set(elems[1:])
    state.nodes[elems[0]] = deps
    for dep in deps:
        if dep & 1:
            state.nodes[dep] = {dep ^ 1}
    return True


SYM_PATTERN = re.compile(r"([ilo])(\d+) (.*)\n")


def parse_symbol(state, line) -> bool:
    match = SYM_PATTERN.match(line)
    if match is None:
        return False

    kind, idx, name = match.groups()

    table = {
        'i': state.symbols.inputs,
        'o': state.symbols.outputs,
        'l': state.symbols.latches
    }.get(kind)
    table[int(idx)] = name
    return True


def parse_comment(state, line) -> bool:
    if state.comments is not None:
        state.comments.append(line.rstrip())
    elif line.rstrip() == 'c':
        state.comments = []
    else:
        return False
    return True


def parse_seq():
    yield parse_header
    yield parse_input
    yield parse_latch
    yield parse_output
    yield parse_and
    yield parse_symbol
    yield parse_comment


def finish_table(table, keys):
    assert len(table) <= len(keys)
    return bidict({table[i]: key for i, key in enumerate(keys)})


def parse(lines, to_aig: bool = True):
    if isinstance(lines, str):
        lines = io.StringIO(lines)

    state = State()
    parsers = parse_seq()
    parser = next(parsers)

    for line in lines:
        while not parser(state, line):
            parser = next(parsers)

            if parser is None:
                raise ValueError(NOT_DONE_PARSING_ERROR.format(state))

    if parser not in (parse_comment, parse_symbol):
        raise ValueError(DONE_PARSING_ERROR.format(state))

    assert state.header.num_ands == 0
    assert state.remaining_inputs == 0
    assert state.remaining_outputs == 0
    assert state.remaining_latches == 0

    # Complete Symbol Table.
    inputs = bidict(finish_table(state.symbols.inputs, state.inputs))
    outputs = finish_table(state.symbols.outputs, state.outputs)
    latches = bidict(finish_table(state.symbols.latches, state.latches))

    # Create expression DAG.
    latch_ids = {latch.id: name for name, latch in latches.items()}
    lit2expr = {}

    for lit in toposort_flatten(state.nodes):
        if lit == 0:
            lit2expr[lit] = A.aig.ConstFalse()
        elif lit in state.inputs:
            lit2expr[lit] = A.aig.Input(inputs.inv[lit])
        elif lit in latch_ids:
            name = latch_ids[lit]
            lit2expr[lit] = A.aig.LatchIn(name)
        elif lit & 1:
            lit2expr[lit] = A.aig.Inverter(lit2expr[lit & -2])
        else:
            nodes = [lit2expr[lit2] for lit2 in state.nodes[lit]]
            lit2expr[lit] = reduce(A.aig.AndGate, nodes)

    return A.aig.AIG(
        inputs=set(inputs),
        node_map={n: lit2expr[lit] for n, lit in outputs.items()},
        latch_map={n: lit2expr[latch.input] for n, latch in latches.items()},
        latch2init={n: latch.init for n, latch in latches.items()},
        comments=tuple(state.comments if state.comments else []),
    )


def load(path: str, to_aig: bool = True):
    with open(path, 'r') as f:
        return parse(''.join(f.readlines()), to_aig=to_aig)


__all__ = ['load', 'parse']



# flake8: noqa
from aiger.aig import AIG, to_aig
from aiger.common import and_gate, or_gate, bit_flipper, source, sink
from aiger.common import tee, empty, identity, delay, parity_gate
from aiger.parser import *
from aiger.expr import atom, atoms, ite, BoolExpr
from aiger.lazy import *



import funcy as fn

import attr
from bidict import bidict
from sortedcontainers import SortedDict


AAG_HEADER = "aag {} {} {} {} {}\n"


def header(inputs, outputs, latchins, latchouts, inits, count):
    n_in, n_lin, n_out = len(inputs), len(latchins), len(outputs)
    n_and = count - n_in - n_lin
    assert len(latchouts) == n_lin == len(inits)

    buff = AAG_HEADER.format(count, n_in, n_lin, n_out, n_and)

    # Note: lits sorted for stable output order.
    if inputs:
        buff += "\n".join(map(str, inputs.values())) + "\n"
    for key in sorted(latchins, key=latchins.get):
        buff += f"{latchins[key]} {latchouts[key]} {int(inits[key])}\n"
    if outputs:
        buff += "\n".join(map(str, sorted(outputs.values()))) + "\n"
    return buff


def footer(inputs, latches, outputs, comments):
    # Note: relies on the fact that dictionaries respect insertion order.
    buff = ""
    for pre, elems in zip("iol", (inputs, outputs, latches)):
        if not elems:
            continue
        buff += "\n".join(f"{pre}{i} {k}" for i, k in enumerate(elems)) + "\n"
    if comments:
        buff += "c\n"
        buff += "\n".join(comments) + "\n"
    return buff


def dump(circ):
    # Create Algebra to write to a string buffer.
    buff = ""
    count = 0

    @attr.s(auto_attribs=True, frozen=True, repr=False)
    class NodeAlg:
        lit: int

        def __repr__(self):
            return str(self.lit)

        @fn.memoize
        def __and__(self, other):
            nonlocal count
            nonlocal buff

            count += 1
            new = NodeAlg(count << 1)

            right, left = sorted([self.lit, other.lit])
            buff += f"{new} {left} {right}\n"
            return new

        def __invert__(self):
            return NodeAlg(self.lit ^ 1)

    def lift(obj) -> NodeAlg:
        if isinstance(obj, bool):
            return NodeAlg(int(obj))
        elif isinstance(obj, NodeAlg):
            return obj
        raise NotImplementedError

    # Inputs and Latches are first indices.

    start = count + 1
    inputs = bidict(
        {k: NodeAlg(i << 1) for i, k in enumerate(sorted(circ.inputs), start)}
    )
    count += len(inputs)
    start = count + 1
    latches = bidict(
        {k: NodeAlg(i << 1) for i, k in enumerate(sorted(circ.latches), start)}
    )
    count += len(latches)

    # Interpret circ over Algebra.
    omap, lmap = circ(inputs=inputs, latches=latches, lift=lift)
    omap = SortedDict(omap.get, omap)  # Force outputs to be ordered.

    # Add header and footer.
    head = header(inputs, omap, latches, lmap, circ.latch2init, count)
    foot = footer(inputs, latches, omap, circ.comments)

    return head + buff + foot


__all__ = ['dump']



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiger']

package_data = \
{'': ['*']}

install_requires = \
['attrs>=21,<22',
 'bidict>=0.21.0,<0.22.0',
 'funcy>=1.12,<2.0',
 'parsimonious>=0.8.1,<0.9.0',
 'pyrsistent>=0.18.0,<0.19.0',
 'sortedcontainers>=2.3.0,<3.0.0',
 'toposort>=1.5,<2.0']

setup_kwargs = {
    'name': 'py-aiger',
    'version': '6.1.19',
    'description': 'A python library for manipulating sequential and-inverter gates.',
    'long_description': '<figure>\n  <img src="assets/logo_text.svg" alt="py-aiger logo" width=300px>\n  <figcaption>\n      pyAiger: A python library for manipulating sequential and\n      combinatorial circuits.\n  </figcaption>\n\n</figure>\n\n\n[![Build Status](https://cloud.drone.io/api/badges/mvcisback/py-aiger/status.svg)](https://cloud.drone.io/mvcisback/py-aiger)\n[![codecov](https://codecov.io/gh/mvcisback/py-aiger/branch/master/graph/badge.svg)](https://codecov.io/gh/mvcisback/py-aiger)\n[![PyPI version](https://badge.fury.io/py/py-aiger.svg)](https://badge.fury.io/py/py-aiger)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![DOI](https://zenodo.org/badge/DOI/10.5281/zenodo.1405781.svg)](https://doi.org/10.5281/zenodo.1405781)\n\n<!-- markdown-toc start - Don\'t edit this section. Run M-x markdown-toc-generate-toc again -->\n**Table of Contents**\n\n- [About PyAiger](#about-pyaiger)\n- [Installation](#installation)\n- [Boolean Expression DSL](#boolean-expression-dsl)\n- [Sequential Circuit DSL](#sequential-circuit-dsl)\n    - [Sequential composition](#sequential-composition)\n    - [Parallel composition](#parallel-composition)\n    - [Circuits with Latches/Feedback/Delay](#circuits-with-latchesfeedbackdelay)\n    - [Relabeling](#relabeling)\n    - [Evaluation](#evaluation)\n    - [Useful circuits](#useful-circuits)\n- [Extra](#extra)\n- [Ecosystem](#ecosystem)\n- [Related Projects](#related-projects)\n- [Citing](#citing)\n\n<!-- markdown-toc end -->\n\n\n# About PyAiger\n\n1. Q: How is Py-Aiger pronounced? A: Like "pie" + "grrr".\n2. Q: Why python? Aren\'t you worried about performance?! A: No. The goals of this library are ease of use and hackability. \n3. Q: No, I\'m really concerned about performance! A: This library is not suited to implement logic solvers. For everything else, such as the creation and manipulation of circuits with many thousands of gates in between solver calls, the library is really fast enough.\n4. Q: Where does the name come from? A: <a href="http://fmv.jku.at/aiger/">Aiger</a> is a popular circuit format. The format is used in <a href="http://fmv.jku.at/hwmcc17/">hardware model checking</a>, <a href="http://www.syntcomp.org/">synthesis</a>, and is supported by <a href="https://github.com/berkeley-abc/abc">ABC</a>. The name is a combination of AIG (standing for <a href="https://en.wikipedia.org/wiki/And-inverter_graph">And-Inverter-Graph</a>) and the mountain <a href="https://en.wikipedia.org/wiki/Eiger">Eiger</a>.\n\n# Installation\n\nIf you just need to use `aiger`, you can just run:\n\n`$ pip install py-aiger`\n\nFor developers, note that this project uses the\n[poetry](https://poetry.eustace.io/) python package/dependency\nmanagement tool. Please familarize yourself with it and then\nrun:\n\n`$ poetry install`\n\n# Usage\n\n```\nimport aiger\n\nx, y, z, w = aiger.atoms(\'x\', \'y\', \'z\', \'w\')\n\nexpr1 = z.implies(x & y)\nexpr2 = z & w\n\ncirc1 = expr1.with_output(\'z\') \\      # Get AIG for expr1 with output \'z\'.\n             .aig\ncirc2 = circ1 >> circ2                # Feed outputs of circ1 into circ2.\n```\n\n# Boolean Expression DSL\nWhile powerful, when writing combinatorial circuits, the Sequential\nCircuit DSL can be somewhat clumsy. For this common usecase, we have\ndeveloped the Boolean Expression DSL. All circuits generated this way\nhave a single output.\n\n```python\nimport aiger\nx, y, z = aiger.atoms(\'x\', \'y\', \'z\')\nexpr1 = x & y  # circuit with inputs \'x\', \'y\' and 1 output computing x AND y.\nexpr2 = x | y  # logical or.\nexpr3 = x ^ y  # logical xor.\nexpr4 = x == y  # logical ==, xnor.\nexpr5 = x.implies(y)\nexpr6 = ~x  # logical negation.\nexpr7 = aiger.ite(x, y, z)  # if x then y else z.\n\n# Atoms can be constants.\nexpr8 = x & True  # Equivalent to just x.\nexpr9 = x & False # Equivalent to const False.\n\n# Specifying output name of boolean expression.\n# - Output is a long uuid otherwise.\nexpr10 = expr5.with_output(\'x_implies_y\')\nassert expr10.output == \'x_implies_y\'\n\n# And you can inspect the AIG if needed.\ncirc = x.aig\n\n# And of course, you can get a BoolExpr from a single output aig.\nexpr10 = aiger.BoolExpr(circ)\n```\n\n\n# Sequential Circuit DSL\n\n```python\nimport aiger\nfrom aiger import utils\n\n# Parser for ascii AIGER format.\naig1 = aiger.load(path_to_aig1_file.aag)\naig2 = aiger.load(path_to_aig2_file.aag)\n```\n\n## Sequential composition\n```python\naig3 = aig1 >> aig2\n```\n\n## Parallel composition\n```python\naig4 = aig1 | aig2\n```\n\n## Circuits with Latches and Delayed Feedback\nSometimes one requires some outputs to be feed back into the circuits\nas time delayed inputs. This can be accomplished using the `loopback`\nmethod. This method takes in a variable number of dictionaries\nencoding how to wire an input to an output. The wiring dictionaries\nwith the following keys and default values:\n\n| Key         | Default | Meaning                          |\n| ----------- | ------- | -------------------------------- |\n| input       |         | Input port                       |\n| output      |         | Output port                      |\n| latch       | input   | Latch name                       |\n| init        | True    | Initial latch value              |\n| keep_output | True    | Keep loopbacked output as output |\n\n\n```python\n# Connect output y to input x with delay, initialized to True.\n# (Default initialization is False.)\naig5 = aig1.loopback({\n  "input": "x", "output": "y",\n})\n\naig6 = aig1.loopback({\n  "input": "x", "output": "y",\n}, {\n  "input": "z", "output": "z", "latch": "z_latch",\n  "init": False, "keep_outputs": False\n})\n\n```\n\n## Relabeling\n\nThere are two syntaxes for relabeling. The first uses indexing\nsyntax in a nod to [notation often used variable substition in math](https://mathoverflow.net/questions/243084/history-of-the-notation-for-substitution).\n\nThe syntax is the relabel method, which is preferable when one wants\nto be explicit, even for those not familar with `py-aiger`.\n\n```python\n# Relabel input \'x\' to \'z\'.\naig1[\'i\', {\'x\': \'z\'}]\naig1.relabel(\'input\', {\'x\': \'z\'})\n\n# Relabel output \'y\' to \'w\'.\naig1[\'o\', {\'y\': \'w\'}]\naig1.relabel(\'output\', {\'y\': \'w\'})\n\n# Relabel latches \'l1\' to \'l2\'.\naig1[\'l\', {\'l1\': \'l2\'}]\naig1.relabel(\'latch\', {\'l1\': \'l2\'})\n```\n\n## Evaluation\n```python\n# Combinatoric evaluation.\naig3(inputs={\'x\':True, \'y\':False})\n\n# Sequential evaluation.\nsim = aig3.simulate([{\'x\': 0, \'y\': 0}, \n                    {\'x\': 1, \'y\': 2},\n                    {\'x\': 3, \'y\': 4}])\n\n# Simulation Coroutine\nsim = aig3.simulator()  # Coroutine\nnext(sim)  # Initialize\nprint(sim.send({\'x\': 0, \'y\': 0}))\nprint(sim.send({\'x\': 1, \'y\': 2}))\nprint(sim.send({\'x\': 3, \'y\': 4}))\n\n\n# Unroll\naig4 = aig3.unroll(steps=10, init=True)\n```\n\n## Useful circuits\n```python\n# Fix input x to be False.\naig4 = aiger.source({\'x\': False}) >> aig3\n\n# Remove output y. \naig4 = aig3 >> aiger.sink([\'y\'])\n\n# Create duplicate w of output y.\naig4 = aig3 >> aiger.tee({\'y\': [\'y\', \'w\']})\n```\n\n# Extra\n```python\naiger.common.eval_order(aig1)  # Returns topological ordering of circuit gates.\n\n# Convert object into an AIG from multiple formats including BoolExpr, AIG, str, and filepaths.\naiger.to_aig(aig1)\n```\n\n# Ecosystem\n\n### Stable\n- [py-aiger-bv](https://github.com/mvcisback/py-aiger-bv): Extension of pyAiger for manipulating sequential bitvector circuits.\n- [py-aiger-cnf](https://github.com/mvcisback/py-aiger-cnf): BoolExpr to Object representing CNF. Mostly used for interfacing with py-aiger-sat.\n- [py-aiger-past-ltl](https://github.com/mvcisback/py-aiger-past-ltl): Converts Past Linear Temporal Logic to aiger circuits.\n- [py-aiger-coins](https://github.com/mvcisback/py-aiger-coins): Library for creating circuits that encode discrete distributions.\n- [py-aiger-sat](https://github.com/mvcisback/py-aiger-sat): Bridge between py-aiger and py-sat.\n- [py-aiger-bdd](https://github.com/mvcisback/py-aiger-bdd): Aiger <-> BDD bridge.\n- [py-aiger-gridworld](https://github.com/mvcisback/py-aiger-gridworld): Create aiger circuits representing gridworlds.\n- [py-aiger-dfa](https://pypi.org/project/py-aiger-dfa/): Convert between finite automata and sequential circuits.\n\n\n### Underdevelopment\n\n\n- [py-aiger-spectral](https://github.com/mvcisback/py-aiger-spectral): A tool for performing (Fourier) Analysis of Boolean Functions.\n- [py-aiger-abc](https://pypi.org/project/py-aiger-abc/): Aiger and abc bridge.\n\n# Related Projects\n- [pyAig](https://github.com/sterin/pyaig): Another python library\n  for working with AIGER circuits.\n',
    'author': 'Marcell Vazquez-Chanlatte',
    'author_email': 'marcell.vc@eecs.berkeley.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mvcisback/py-aiger',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""
Abstractions for compositions/manipulations of And Inverter Graphs.
"""

from __future__ import annotations

from abc import ABCMeta, abstractmethod
import operator as op
import pathlib
from typing import Tuple, FrozenSet

import attr
import funcy as fn
from pyrsistent import pmap
from pyrsistent.typing import PMap

import aiger as A
from aiger import common as cmn
from aiger import parser
from aiger import writer


@attr.frozen
class Node(metaclass=ABCMeta):
    def __and__(self, other: Node) -> Node:
        if self.is_false or other.is_false:
            return ConstFalse()
        elif self.is_true:
            return other
        elif other.is_true:
            return self
        return AndGate(self, other)

    def __invert__(self) -> Node:
        if isinstance(self, Inverter):
            return self.input
        return Inverter(self)

    @property
    def is_false(self):
        return isinstance(self, ConstFalse)

    @property
    def is_true(self):
        return (~self).is_false

    @property
    @abstractmethod
    def children(self):
        pass


@attr.frozen(cache_hash=True)
class AndGate(Node):
    left: Node
    right: Node

    @property
    def children(self):
        return (self.left, self.right)



@attr.frozen
class Inverter(Node):
    input: Node

    @property
    def children(self):
        return (self.input, )


@attr.frozen
class Input(Node):
    name: str

    @property
    def children(self):
        return ()


@attr.frozen
class LatchIn(Node):
    name: str

    @property
    def children(self):
        return ()


@attr.frozen
class ConstFalse(Node):
    @property
    def children(self):
        return ()

    def __hash__(self):
        return hash(False)


@attr.frozen(repr=False)
class AIG:
    inputs: FrozenSet[str] = frozenset()
    node_map: PMap[str, Node] = attr.ib(default=pmap(), converter=pmap)
    latch_map: PMap[str, Node] = attr.ib(default=pmap(), converter=pmap)
    latch2init: PMap[str, bool] = attr.ib(default=pmap(), converter=pmap)
    comments: Tuple[str] = ()

    def __repr__(self):
        return writer.dump(self)

    def __getitem__(self, others):
        return self.lazy_aig[others].aig

    def __iter_nodes__(self):
        """Returns an iterator over iterators of nodes in an AIG.

        If completely flattened this iterator would give topological
        a order on the nodes, starting on the inputs.

        The reason for the iterator over iterators is to mark
        dependencies. Namely, to compute the value of any node
        requires just the value of the nodes in the previous iterator.
        """
        return [cmn.dfs(self)]

    def evolve(self, **kwargs):
        return attr.evolve(self, **kwargs)

    def relabel(self, key, mapping):
        assert key in {'input', 'output', 'latch'}
        return self[key[0], mapping]

    @property
    def aig(self):
        return self

    @property
    def lazy_aig(self):
        return A.lazy(self)

    @property
    def outputs(self):
        return frozenset(self.node_map.keys())

    @property
    def latches(self):
        return frozenset(self.latch2init.keys())

    @property
    def cones(self):
        return frozenset(self.node_map.values())

    @property
    def latch_cones(self):
        return frozenset(self.latch_map.values())

    def __rshift__(self, other):
        return seq_compose(self, other)

    def __lshift__(self, other):
        return seq_compose(other, self)

    def __or__(self, other):
        return par_compose(self, other)

    def __call__(self, inputs, latches=None, *, lift=None):
        """Evaluate AIG on inputs (and latches).
        If `latches` is `None` initial latch value is used.

        `lift` is an optional argument used to interpret constants
        (False, True) in some other Boolean algebra over (&, ~).

        - See py-aiger-bdd and py-aiger-cnf for examples.
        """
        if latches is None:
            latches = dict()

        if lift is None:
            lift = fn.identity
            and_, neg = op.and_, op.not_
        else:
            and_, neg = op.__and__, op.__invert__

        latchins = fn.merge(dict(self.latch2init), latches)
        # Remove latch inputs not used by self.
        latchins = fn.project(latchins, self.latches)

        latch_map = dict(self.latch_map)
        boundary = set(self.node_map.values()) | set(latch_map.values())

        store, prev, mem = {}, set(), {}
        for node_batch in self.__iter_nodes__():
            prev = set(mem.keys()) - prev
            mem = fn.project(mem, prev)  # Forget about unnecessary gates.

            for gate in node_batch:
                if isinstance(gate, Inverter):
                    mem[gate] = neg(mem[gate.input])
                elif isinstance(gate, AndGate):
                    mem[gate] = and_(mem[gate.left], mem[gate.right])
                elif isinstance(gate, Input):
                    mem[gate] = lift(inputs[gate.name])
                elif isinstance(gate, LatchIn):
                    mem[gate] = lift(latchins[gate.name])
                elif isinstance(gate, ConstFalse):
                    mem[gate] = lift(False)

                if gate in boundary:
                    store[gate] = mem[gate]  # Store for eventual output.

        outs = {out: store[gate] for out, gate in self.node_map.items()}
        louts = {out: store[gate] for out, gate in latch_map.items()}
        return outs, louts

    def simulator(self, latches=None):
        inputs = yield
        while True:
            outputs, latches = self(inputs, latches)
            inputs = yield outputs, latches

    def simulate(self, input_seq, latches=None):
        sim = self.simulator()
        next(sim)
        return [sim.send(inputs) for inputs in input_seq]

    def cutlatches(self, latches=None, check_postcondition=True, renamer=None):
        lcirc, lmap = self.lazy_aig.cutlatches(latches, renamer=renamer)
        return lcirc.aig, lmap

    def loopback(self, *wirings):
        return self.lazy_aig.loopback(*wirings).aig

    def feedback(
        self, inputs, outputs, initials=None, latches=None, keep_outputs=False
    ):
        import warnings
        warnings.warn("deprecated", DeprecationWarning)

        def create_wire(val):
            iname, oname, lname, init = val
            return {
                'input': iname, 'output': oname, 'latch': lname, 'init': init,
                'keep_output': keep_outputs
            }

        if initials is None:
            initials = fn.repeat(False)

        if latches is None:
            assert (set(inputs) & self.latches) == set()
            latches = inputs

        vals = zip(inputs, outputs, latches, initials)
        return self.loopback(*map(create_wire, vals))

    def unroll(self, horizon, *, init=True, omit_latches=True,
               only_last_outputs=False):
        return self.lazy_aig.unroll(
            horizon, init=init, omit_latches=omit_latches,
            only_last_outputs=only_last_outputs
        ).aig

    def write(self, path):
        with open(path, 'w') as f:
            f.write(repr(self))

    def reinit(self, latch2init):
        """Update late initial values based on mapping provided."""
        return self.lazy_aig.reinit(latch2init).aig


def par_compose(aig1, aig2, check_precondition=True):
    return A.Parallel(aig1, aig2).aig


def seq_compose(circ1, circ2, *, input_kind=Input):
    return A.Cascading(circ1, circ2).aig


def to_aig(circ, *, allow_lazy=False) -> AIG:
    if isinstance(circ, pathlib.Path) and circ.is_file():
        circ = parser.load(circ)
    elif isinstance(circ, str):
        if circ.startswith('aag '):
            circ = parser.parse(circ)  # Assume it is an AIGER string.
        else:
            circ = parser.load(circ)  # Assume it is a file path.

    if allow_lazy and hasattr(circ, '.lazy_aig'):
        return circ.lazy_aig

    return circ.aig  # Extract AIG from potential wrapper.



from __future__ import annotations

from typing import Union

import attr

import aiger as A
from aiger import common as cmn


@attr.s(frozen=True, slots=True, eq=False, auto_attribs=True, hash=True)
class BoolExpr:
    _aig: A.AIG_Like

    @property
    def aig(self) -> A.AIG:
        return self._aig.aig

    @property
    def lazy_aig(self) -> A.LazyAig:
        return A.lazy(self._aig)

    def __call__(self, inputs, *, lift=None):
        return self.aig(inputs, lift=lift)[0][self.output]

    def __and__(self, other):
        return _binary_gate(cmn.and_gate, self, other)

    def __or__(self, other):
        return _binary_gate(cmn.or_gate, self, other)

    def __xor__(self, other):
        return _binary_gate(cmn.parity_gate, self, other)

    def __invert__(self):
        return type(self)(
            aig=self.aig >> cmn.bit_flipper(self.aig.outputs, [cmn._fresh()])
        )

    def __eq__(self, other):
        return ~(self ^ other)

    def implies(self, other):
        return ~self | other

    @property
    def output(self):
        return list(self.aig.outputs)[0]

    @property
    def inputs(self):
        return self.aig.inputs

    def with_output(self, val):
        if val == self.output:
            return self
        return attr.evolve(self, aig=self.aig['o', {self.output: val}])

    def _fresh_output(self, name=None):
        if name is None:
            name = cmn._fresh()
        return self.with_output(name)


def _binary_gate(gate, expr1, expr2):
    if isinstance(expr2, (bool, int)):
        expr2 = atom(expr2)
    aig = expr1._fresh_output().aig | expr2._fresh_output().aig
    aig >>= gate(inputs=aig.outputs, output=cmn._fresh())
    return type(expr1)(aig=aig)


def ite(test, expr_true, expr_false):
    return test.implies(expr_true) & (~test).implies(expr_false)


def atom(val: Union[str, bool, None]) -> BoolExpr:
    output = cmn._fresh()
    if val is None:
        val = cmn._fresh()

    if isinstance(val, str):
        assert val not in ('True', 'False')
        aig = cmn.identity([val], [output])
    else:
        aig = cmn.source({output: val})

    return BoolExpr(aig)


def atoms(*vals):
    return tuple(map(atom, vals))


__all__ = ["atom", "atoms", "ite", "BoolExpr"]
